#!/usr/bin/env python

"""
This file contains utility functions to enable easy interfacing
between Python and calls to the NUPACK core executables.  It additionally
contains utility scripts for converting structures from dot-paren
notation to pair lists.
"""
from __future__ import print_function
from __future__ import division

import os
import subprocess
import time
import warnings

import numpy as np
from scipy import sparse
import pandas as pd

# ############################################################### #
def energy(sequence, structure, T=37.0, material='rna1995', prefix=None,
           NUPACKHOME=None, delete_files=True):
    """
    Calculate the energy (in units of kcal/mol) of a particular structure for a
    specified nucleic acid sequence.
    
    Arguments:
    sequence -- A string of nucleic acid base codes
    structure -- A string in dot-paren format of the same length as sequence
    
    Keyword Arguments:
    T -- Temperature in degrees Celsius (default 37.0)
    material -- The material parameters file to use (default 'rna1995')
    prefix -- The file prefix to be used when calling the energy executable
    (default None)
    NUPACKHOME -- The file location of the NUPACK installation (default None)
    delete_files -- A flag to remove the prefix.out and prefix.in files (default
    True)
    
    Return:
    The energy of the nucleic acid sequence in the given secondary structure
    """

    # Check all inputs
    nupack_home = check_nupackhome(NUPACKHOME)
    material = check_material(material)
    prefix = check_prefix(prefix, 'energy')
    
    input_file = prefix + '.in'
    output_file = prefix + '.out'

    # Make the input file
    f = open(input_file, 'w')
    f.write('%s\n%s\n' % (sequence, structure))
    f.close()

    # Run NUPACK's energy executable
    args = [nupack_home + '/bin/energy',
            '-T', str(T),
            '-material', material,
            prefix]
    with open(output_file, 'w') as outfile:
        subprocess.check_call(args, stdout=outfile)
        outfile.close()

    # Parse the output
    en = float(np.loadtxt(output_file, comments='%'))

    # Remove files if requested
    if delete_files:
        subprocess.check_call(['rm', '-f', input_file, output_file])
    
    return en
# ############################################################### #


# ############################################################### #
def prob(sequence, structure, T=37.0, material='rna1995', prefix=None,
           NUPACKHOME=None, delete_files=True):
    """
    Calculate the probability of a nucleic acid sequence adopting a particular
    secondary structure.
    
    Arguments:
    sequence -- A string of nucleic acid base codes
    structure -- A string in dot-paren format of the same length as sequence
    
    Keyword Arguments:
    T -- Temperature in degrees Celsius (default 37.0)
    material -- The material parameters file to use (default 'rna1995')
    prefix -- The file prefix to be used when calling the prob executable
    (default None)
    NUPACKHOME -- The file location of the NUPACK installation (default None)
    delete_files -- A flag to remove the prefix.out and prefix.in files (default
    True)
    
    Return:
    The probability of the nucleic acid sequence adopting the given secondary
    structure
    """

    # Check all inputs
    nupack_home = check_nupackhome(NUPACKHOME)
    material = check_material(material)
    prefix = check_prefix(prefix, 'prob')
    
    input_file = prefix + '.in'
    output_file = prefix + '.out'

    # Make the input file
    f = open(input_file, 'w')
    f.write('%s\n%s\n' % (sequence, structure))
    f.close()

    # Run NUPACK's prob executable
    args = [nupack_home + '/bin/prob',
            '-T', str(T),
            '-material', material,
            prefix]
    with open(output_file, 'w') as outfile:
        subprocess.check_call(args, stdout=outfile)
        outfile.close()

    # Parse the output
    pr = float(np.loadtxt(output_file, comments='%'))

    # Remove files if requested
    if delete_files:
        subprocess.check_call(['rm', '-f', input_file, output_file])
    
    return pr
# ############################################################### #

# ############################################################### #
def pfunc(sequence, T=37.0, material='rna1995', prefix=None, NUPACKHOME=None, 
        delete_files=True):
    """
    Compute the partition function and free energy for a nucleic acid sequence.
    
    Argument:
    sequence -- A string of nucleic acid base codes
    
    Keyword Arguments:
    T -- Temperature in degrees Celsius (default 37.0)
    material -- The material parameters file to use (default 'rna1995')
    prefix -- The file prefix to be used when calling the pfunc executable
    (default None)
    NUPACKHOME -- The file location of the NUPACK installation (default None)
    delete_files -- A flag to remove the prefix.in file (default True)
    
    Return:
    The partition function and free energy.
    """

    # Check all inputs
    nupack_home = check_nupackhome(NUPACKHOME)
    material = check_material(material)
    prefix = check_prefix(prefix, 'pfunc')

    input_file = prefix + '.in'

    # Make the input file
    f = open(input_file, 'w')
    f.write(sequence + '\n')
    f.close()

    # Run NUPACK's mfe executable
    args = [nupack_home + '/bin/pfunc',
            '-T', str(T),
            '-material', material,
            prefix]
    output = subprocess.check_output(args)
    
    output = output.decode('utf-8').split("\n")
    # Parse the output
    energy_pfunc = []
    for line in output:
        if len(line) > 0 and line[0] != '%':
            energy_pfunc.append(float(line))
    
    # Remove files if requested
    if delete_files:
        subprocess.check_call(['rm', '-f', input_file])
        
    energy = energy_pfunc[0]
    pfunc = energy_pfunc[1]
    return pfunc, energy 
# ############################################################### #


# ############################################################### #
def count(sequence, T=37.0, material='rna1995', prefix=None, NUPACKHOME=None, 
        delete_files=True):
    """
    Count the number of structures for a nucleic acid sequence.
    
    Argument:
    sequence -- A string of nucleic acid base codes
    
    Keyword Arguments:
    T -- Temperature in degrees Celsius (default 37.0)
    material -- The material parameters file to use (default 'rna1995')
    prefix -- The file prefix to be used when calling the count executable
    (default None)
    NUPACKHOME -- The file location of the NUPACK installation (default None)
    delete_files -- A flag to remove the prefix.in file (default True)
    
    Return:
    A count of the number of unique structures consistent with the input 
    sequence.
    """

    # Check all inputs
    nupack_home = check_nupackhome(NUPACKHOME)
    material = check_material(material)
    prefix = check_prefix(prefix, 'count')

    input_file = prefix + '.in'

    # Make the input file
    f = open(input_file, 'w')
    f.write(sequence + '\n')
    f.close()

    # Run NUPACK's mfe executable
    args = [nupack_home + '/bin/count',
            '-T', str(T),
            '-material', material,
            prefix]
    output = subprocess.check_output(args)
    
    output = output.decode('utf-8').split("\n")
    # Parse the output
    for line in output:
        if len(line) > 0 and line[0] != '%':
            num_structures = float(line)
    
    # Remove files if requested
    if delete_files:
        subprocess.check_call(['rm', '-f', input_file])

    return num_structures
# ############################################################### #


# ############################################################### #
def mfe(sequence, T=37.0, material='rna1995', prefix=None, NUPACKHOME=None, 
        delete_files=True):
    """
    Calculate the MFE and MFE structure of a nucleic acid sequence.
    
    Argument:
    sequence -- A string of nucleic acid base codes
    
    Keyword Arguments:
    T -- Temperature in degrees Celsius (default 37.0)
    material -- The material parameters file to use (default 'rna1995')
    prefix -- The file prefix to be used when calling the mfe executable
    (default None)
    NUPACKHOME -- The file location of the NUPACK installation (default None)
    delete_files -- A flag to remove the prefix.mfe and prefix.in files 
    (default True)
    
    Return:
    The MFE and MFE structure in dot-paren format. If there are multiple MFE
    structures, only one is returned.
    """

    # Check all inputs
    nupack_home = check_nupackhome(NUPACKHOME)
    material = check_material(material)
    prefix = check_prefix(prefix, 'mfe')

    input_file = prefix + '.in'
    output_file = prefix + '.mfe'

    # Make the input file
    f = open(input_file, 'w')
    f.write(sequence + '\n')
    f.close()

    # Run NUPACK's mfe executable
    args = [nupack_home + '/bin/mfe',
            '-T', str(T),
            '-material', material,
            prefix]
    subprocess.check_call(args)

    # Parse the output
    df = pd.read_csv(output_file, header=None, comment='%', names=['col'],
                     error_bad_lines=False)
    en = float(df.col[1])
    struct = df.col[2]
    
    # Remove files if requested
    if delete_files:
        subprocess.check_call(['rm', '-f', input_file, output_file])

    return struct, en
# ############################################################### #


# ############################################################### #
def pairs(sequence, T=37.0, material='rna1995', prefix=None, NUPACKHOME=None, 
          delete_files=True):
    """
    Calculate the probabilities of all possible base pairs of a nucleic acid
    sequence over the ensemble of unpseudoknotted structures.
    
    Argument:
    sequence -- A string of nucleic acid base codes
    
    Keyword Arguments:
    T -- Temperature in degrees Celsius (default 37.0)
    material -- The material parameters file to use (default 'rna1995')
    prefix -- The file prefix to be used when calling the pairs executable
    (default None)
    NUPACKHOME -- The file location of the NUPACK installation (default None)
    delete_files -- A flag to remove the prefix.ppairs and prefix.in files 
    (default True)
    
    Return:
    The pair probabilities matrix right-augmented with the unpaired
    probabilities.
    
    Notes:
    Pseudoknots are not allowed.
    Pair probability matrices are symmetric, by definition, except for the
    unpaired probability column).
    """

    # Check all inputs
    nupack_home = check_nupackhome(NUPACKHOME)
    material = check_material(material)
    prefix = check_prefix(prefix, 'mfe')

    input_file = prefix + '.in'
    output_file = prefix + '.ppairs'

    # Make the input file
    f = open(input_file, 'w')
    f.write(sequence + '\n')
    f.close()

    # Run NUPACK's mfe executable
    args = [nupack_home + '/bin/pairs',
            '-T', str(T),
            '-material', material,
            prefix]
    subprocess.check_call(args)

    # Parse the output
    df = pd.read_csv(output_file, header=None, comment='%', 
                     names=['i', 'j', 'p'], delimiter='\t')

    # Build a sparse pair probabilities matrix with indexing starting at 0
    # df.ix[0] is garbage.  df.i[1] is # of bases.  Indices 2 and up are probs.
    P = sparse.csc_matrix((df.p[1:], (df.i[1:] - 1, df.j[1:] - 1)), 
                          shape=(int(df.i[0]), int(df.i[0])+1))
    
    # Fill in lower triangle (ignore sparse efficiency warning)
    # Note: don't have to worry about diagonal; it's necessarily zero
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        P[:,:-1] = P[:,:-1] + P[:,:-1].transpose()

    # Remove files if requested
    if delete_files:
        subprocess.check_call(['rm', '-f', input_file, output_file])

    return np.array(P.todense())
# ############################################################### #


# ############################################################### #
def dotparens_2_pairlist(structure):
    """
    Convert a dot-paren structure into a list of pairs called pairlist
    and an array, plist, whose entry at index i is the base paired to base i.  
    
    Argument:
    structure -- A string in dot-paren format
    
    Return:
    pairlist -- An array of ordered pairs for each base pair
    plist -- An array such that plist[i] is paired with i
    
    Notes:
    Array indexing in Python is zero-based.
    plist[i] is -1 if base i is unpaired.
    This only works for single-stranded structures, and not complexes with
    multiple strands.
    
    Example:
    structure = '.(((...)))'
    
    pairlist = np.array([[1,9],[2,8],[3,7]]) 
    plist = np.array([-1,9,8,7,-1,-1,-1,3,2,1])
    """

    # Length of the sequence
    seqlen = len(structure)
    
    pairlist = []
    leftlist = []
    ind = 0
    # While loop steps through list.  Each left bracket is stored.
    # Whenever we get a right bracket, it is necessarily pair with
    # the last left bracket in leftlist.  This pair is documented
    # and the first entry in leftlist is then deleted.
    while ind < seqlen:
        if structure[ind] == '(':
            leftlist.append(ind)
        elif structure[ind] == ')':
            pairlist.append([leftlist[-1],ind])
            leftlist.remove(leftlist[-1])
        ind = ind + 1

    pairlist.sort()

    # Get plist
    plist = [-1]*seqlen
    for x in pairlist:
        plist[x[0]] = x[1]
        plist[x[1]] = x[0]

    return np.array(pairlist), np.array(plist)
# ############################################################### #


# ############################################################### #
def check_prefix(prefix=None, calc_style=''):
    """
    Generate a prefix for a file for a NUPACK calculation.
    """

    # If prefix is provided, make sure file does not already exist
    if prefix is not None:
        if os.path.isfile(prefix + '.in') or os.path.isfile(prefix + '.out') \
           or os.path.isfile(prefix + '.mfe') \
           or os.path.isfile(prefix + '.ppairs'):
            raise ValueError('Files with specified prefix already exist.')
        else:
            return prefix
    else:
        # Generate prefix based on time.
        prefix = time.strftime('%Y-%m-%d-%H_%M_%S_')
        prefix += str(calc_style)
        
        # Check to make sure file name does not already exist
        prefix_base = prefix
        i = 0
        while os.path.isfile(prefix + '.in') \
              or os.path.isfile(prefix + '.out') \
              or os.path.isfile(prefix + '.mfe') \
              or os.path.isfile(prefix + '.ppairs'):
            prefix = prefix_base + '_%08d' % i
            i += 1

        return prefix
    
# ############################################################### #
def check_nupackhome(user_supplied_dir=None):
    """
    Validate or generate a string with the NUPACKHOME directory path.
    
    Notes:
    If user_supplied_dir is not None, checks to make sure the directory looks
    like NUPACKHOME and then strips trailing slash if there is one.
    """
    

    if user_supplied_dir is None:
        try:
            nupackhome = os.environ['NUPACKHOME']
        except KeyError:
            raise RuntimeError('NUPACKHOME environment variable not set.')
    elif user_supplied_dir.endswith('/'):
        # Strip trailing slash if there is one
        nupackhome = user_supplied_dir[:-1]

    # Make sure NUPACK looks ok and has executables
    if os.path.isdir(nupackhome + '/bin') \
       and os.path.isfile(nupackhome + '/bin/mfe') \
       and os.path.isfile(nupackhome + '/bin/pairs') \
       and os.path.isfile(nupackhome + '/bin/energy') \
       and os.path.isfile(nupackhome + '/bin/prob'):
        return nupackhome
    else:
        raise RuntimeError('NUPACK not compiled in %s.' % nupackhome)
            
# ############################################################### #
def check_material(material):
    """
    Check material input.

    Notes:
    The 'rna1999' parameters will not work with temperatures other than 37.0
    degrees C.
    """

    if material not in ['rna1999', 'dna1998', 'rna1995']:
        print('!! Improper material parameter. Allowed values are:')
        print('!!   ''rna1999'', ''rna1995'', ''dna1998''.')
        raise ValueError('Improper material parameter.')
    else:
        return material





